import { world, system } from "@minecraft/server";

let savedAreas = new Map();
let protectedAreas = new Map();
let initialDroppedItems = new Map();
let playerDroppedItems = new Set();
let itemRemovalIntervalId;

function shouldScriptRun() {
    return world.getAllPlayers().some(player => player.hasTag("ger_user"));
}

world.afterEvents.entitySpawn.subscribe((event) => {
    const { entity } = event;
    if (entity.typeId !== "minecraft:item") return;

    const item = entity.getComponent("item").itemStack;

    const player = world.getPlayers().find(p => {
        return entity.dimension.getEntities({
            type: "minecraft:player",
            location: entity.location,
            maxDistance: 0.05,
        }).includes(p);
    });

    if (player) {
        playerDroppedItems.add(entity.id);
    }
});

function saveArea(player) {
    if (!shouldScriptRun()) return;

    const pos = player.location;
    const radius = 10;
    const maxEntityDistance = 100;

    let entityData = new Map();
    let areaData = [];
    let protectedBlocks = new Set();
    let existingItems = new Set();

    player.removeTag("rt0");

    let allEntities = player.dimension.getEntities();
    for (let entity of allEntities) {
        if (entity.typeId === "minecraft:item" || entity.typeId === "minecraft:xp_orb") continue;
        if (entity.typeId === "minecraft:player" && entity.hasTag("ger_user")) continue;
        if (entity.hasTag("stand")) continue;

        const distance = Math.sqrt(
            Math.pow(entity.location.x - pos.x, 2) +
            Math.pow(entity.location.y - pos.y, 2) +
            Math.pow(entity.location.z - pos.z, 2)
        );

        if (distance <= maxEntityDistance) {
            entityData.set(entity.id, {
                entity: entity,
                startPos: { ...entity.location },
                path: [],
                tagged: false
            });
            entity.addTag("rt22");
        }
    }

    for (let x = pos.x - radius; x <= pos.x + radius; x++) {
        for (let y = Math.max(pos.y - radius, -64); y <= Math.min(pos.y + radius, 319); y++) {
            for (let z = pos.z - radius; z <= pos.z + radius; z++) {
                let block = player.dimension.getBlock({ x, y, z });
                if (block) {
                    let posKey = `${x},${y},${z}`;
                    let blockPermutation = block.permutation;
                    let blockTypeId = block.typeId;
                    let blockStates = blockPermutation.getAllStates();

                    areaData.push({
                        pos: { x, y, z },
                        typeId: blockTypeId,
                        states: blockStates
                    });
                    protectedBlocks.add(posKey);
                }
            }
        }
    }

    savedAreas.set(player.name, { entityData, areaData, playerStartPos: { ...player.location }, playerTagged: false });
    protectedAreas.set(player.name, protectedBlocks);
    initialDroppedItems.set(player.name, existingItems);

    system.runTimeout(() => removeNewDroppedItems(player), 5);
    player.runCommandAsync("execute at @a[tag=ger_user] run function rtzreturne1");
    system.runTimeout(() => trackEntityMovement(player), 5);
}

function trackEntityMovement(player) {
    if (!shouldScriptRun()) return;

    let savedData = savedAreas.get(player.name);
    if (!savedData) return;

    let { entityData } = savedData;

    let intervalId = system.runInterval(() => {
        for (let [entityId, entityInfo] of entityData) {
            let entity = world.getEntity(entityId);
            if (entity) {
                let lastPos = entityInfo.path.length > 0 ? entityInfo.path[entityInfo.path.length - 1] : null;
                let currentPos = { ...entity.location };

                if (!lastPos || lastPos.x !== currentPos.x || lastPos.y !== currentPos.y || lastPos.z !== currentPos.z) {
                    entityInfo.path.push(currentPos);
                }
            }
        }
    }, 5);
}

function restoreBlocks(player) {
    if (!shouldScriptRun()) return;

    let savedData = savedAreas.get(player.name);
    if (!savedData) return;

    let { areaData } = savedData;

    player.runCommandAsync("execute at @a[tag=ger_user] run function rtzreturne2");

    function processRestoreBlocks(index) {
        let end = Math.min(index + 250, areaData.length);
        for (let i = index; i < end; i++) {
            let blockData = areaData[i];
            let block = player.dimension.getBlock(blockData.pos);
            if (block) {
                block.setType(blockData.typeId);
                let blockPermutation = block.permutation;
                for (let [key, value] of Object.entries(blockData.states)) {
                    blockPermutation = blockPermutation.withState(key, value);
                }
                block.setPermutation(blockPermutation);
            }
        }

        if (end < areaData.length) {
            system.runTimeout(() => processRestoreBlocks(end), 1);
        } else {
            player.runCommandAsync("execute at @a[tag=ger_user] run function rtzreturne3");
            system.runTimeout(() => moveEntitiesBack(player), 1);
        }
    }

    processRestoreBlocks(0);
}

function moveEntitiesBack(player) {
    if (!shouldScriptRun()) return;

    let savedData = savedAreas.get(player.name);
    if (!savedData) return;

    let { entityData, playerStartPos, playerTagged } = savedData;
    let startTime = Date.now();
    let attempts = 0;

    let intervalId = system.runInterval(() => {
        if (attempts === 0) {
            player.runCommandAsync("execute at @a[tag=ger_user] run function rtzreturne");
        }

        let allReached = true;
        let currentTime = Date.now();
        attempts++;

        for (let [entityId, entityInfo] of entityData) {
            let entity = world.getEntity(entityId);
            if (!entity || entityInfo.path.length === 0) continue;

            if (entity.hasTag("stand") || entity.hasTag("ger_user")) continue;

            if (!entityInfo.tagged) {
                entity.addTag("rt2");
                entityInfo.tagged = true;
            }

            if (currentTime - startTime >= 14000 || attempts >= 300) {
                entity.teleport(entityInfo.startPos, { checkForBlocks: false });
                entityInfo.path = [];
                entity.removeTag("rt2");
                continue;
            }

            let nextPos = entityInfo.path.pop();
            if (nextPos) {
                entity.teleport(nextPos, { checkForBlocks: false });
                allReached = false;
            } else {
                entity.removeTag("rt2");
            }
        }

        if (player && playerStartPos && !player.hasTag("ger_user")) {
            if (!playerTagged) {
                player.addTag("rt2");
                savedData.playerTagged = true;
            }
            player.teleport(playerStartPos, { checkForBlocks: false });
            player.removeTag("rt2");
        }

        if (allReached || attempts >= 300) {
            system.clearRun(intervalId);
            savedAreas.delete(player.name);
            protectedAreas.delete(player.name);
            initialDroppedItems.delete(player.name);
            player.runCommandAsync("execute at @a[tag=ger_user] run function ger000");

            system.clearRun(itemRemovalIntervalId);
        }
    }, 2);
}

function removeNewDroppedItems(player) {
    if (!shouldScriptRun()) return;

    let savedInitialItems = initialDroppedItems.get(player.name);
    if (!savedInitialItems) return;

    let savedData = savedAreas.get(player.name);
    if (!savedData) return;

    let { playerStartPos } = savedData;
    let areaRadius = 10;

    itemRemovalIntervalId = system.runInterval(() => {
        let allEntities = player.dimension.getEntities();

        for (let entity of allEntities) {
            if (entity.typeId === "minecraft:item" && !savedInitialItems.has(entity.id) && !playerDroppedItems.has(entity.id)) {

                const distance = Math.sqrt(
                    Math.pow(entity.location.x - playerStartPos.x, 2) +
                    Math.pow(entity.location.y - playerStartPos.y, 2) +
                    Math.pow(entity.location.z - playerStartPos.z, 2)
                );

                if (distance <= areaRadius) {
                    entity.kill();
                }
            }
        }
    }, 5);
}

system.runInterval(() => {
    if (!shouldScriptRun()) return;

    for (const player of world.getAllPlayers()) {
        if (player.hasTag("rt0") && !savedAreas.has(player.name)) {
            saveArea(player);
            player.runCommandAsync("execute at @a[tag=ger_user] run function ger00");
            system.runTimeout(() => restoreBlocks(player), 300);
        }
    }
}, 10);
